	TITLE Upload ROM 1.3
**miscdoc+**************************************************
*
*     File: UpLoad ROM
*  Machine: Enterprise
*  Created: November/14/2000 by C. Giesselink
* Modified: July/03/2001, C. Giesselink: reset TCS register
*           July/21/2022, C. Giesselink: wait for
*           transmitter empty before closing serial port
*           and added CRC support
*           November/15/2023, C. Giesselink: after last
*           block send, wait for ACK before sending EOT
*  Purpose: Upload the 1MB ROM content to a PC
*
* More information can be found there:
* https://hp.giesselink.com/
*
* To compile it do the following:
* >@echo HP39AscA A 11 HP39DIR.0004 Root>HP39DIR.CUR
* >@echo HP39AscA B 10 ROMUPL.00010 Upload ROMH 12 LIB1537L.0008 LIB1537L>HP39DIR.000
* >
* >@echo TITLE    Upload ROM>ROMUPL.M
* >@echo OUTPUT   RomUpl.000>>ROMUPL.M
* >@echo LLIST    RomUpl.lr>>ROMUPL.M
* >@echo SUPPRESS XR>>ROMUPL.M
* >@echo OPTION   CODE>>ROMUPL.M
* >@echo REL      RomUpl.o>>ROMUPL.M
* >@echo END>>ROMUPL.M
* >
* >rplcomp romupl.s romupl.a
* >sasm -B -W -a romupl.lst -o romupl.o romupl.a
* >sload -H romupl.m
*
* Then start the HP39G connectivity kit and download the
* aplet to your HP39/40G
**+miscdoc**************************************************

INCLUDE =Entry39.h
INCLUDE =Common.h

ASSEMBLE

=ANNCTRL	EQU	#0010B
=BAU		EQU	#0010D
=IOC		EQU     #00110
=RCS		EQU     #00111
=TCS		EQU     #00112
=RBR		EQU     #00114
=TBR		EQU     #00116
=IRC		EQU	#0011A

=CurROMBank2	EQU	#860B8	saved BS address

* flow control
SOH	EQU	#01
EOT	EQU	#04
ACK	EQU	#06
NAK	EQU	#15
CAN	EQU	#18
CHAR_C	EQU	#43

* bit definitions
sCRC	EQU	0		CRC mode
sFirst	EQU	1		first NAK
sCAN	EQU	2		first CAN
sEOT	EQU	3		finished, send EOT

ASSEMBLE
	Dir_Head 10,Upload ROM,601

RPL

     StndtType		( This aplet skeleton will only provide access to the "C" type of aplet )

INCLUDE =Sharedvar.h	( Include all the variables needed by the aplet structure )

ASSEMBLE
	Vfield  L34,IntTable
	CON(5)	=DOHSTR
	REL(5)  endLink
	Link	_Main
	Link	_Entry
	Link	_Action

endLink
RPL

DEFINE Main		INT_00
DEFINE Entry		INT_01
DEFINE Action		INT_02

ASSEMBLE
	Vfield  IntTable,DirCode
RPL
ASSEMBLE
	CON(5)	=DOHSTR
	REL(5)  endCode
RPL

*
* Main code entry
*
NAMELESS _Main
  ::

     StndCheck			( Default handler for the aplet )

     ' Entry			( view entry )
     ' LeaveGraphView		( view exit )

     '				( view display )
	::
	EnsureMenuOff		( **Remove this line if you want the menu to be displayed )
	;
     '				( view hard key handlers )
	::
	  {
	   { kcEnter :: TakeOver Action ; }	( **ENTER will start your main code )
	   NULL{}
	   NULL{}
	   NULL{}
	   NULL{}
	   NULL{}
          }
          ONE KeyFace
       ;

    TRUE				( **Allow normal keys** )

    NULL{}				( *No menu* )

    ONE 				( **Initial menu page** )

    '					( **Error object** )
    ::	TURNMENUON RECLAIMDISP		( Ensure AGROB to be displayed when an error occurs )
	ERRJMP
    ;
  ;

NAMELESS _Entry ( -> ) ( Display a short message )
  ::
	TURNMENUOFF
	HARDBUFF
	ZEROZERO BINT_131d SIXTYFOUR
	GROB!ZERODRP

	NULL$ DUP
	$ "with ENTER..."
	$ "Start ROM Upload"
	NULL$ DUPDUP

	SEVEN ZERO
	DO
	 FOUR EIGHT INDEX@ #*
	 Put5x7
	LOOP
  ;

NAMELESS _Action ( Put your code here )
  ::
  	( clear display)
	HARDBUFF
	ZEROZERO BINT_131d SIXTYFOUR
	GROB!ZERODRP

	$ "Uploading..."
	FOUR TWENTYFOUR Put5x7

CODE
	GOSBVL	=SAVPTR
	GOSBVL	=DisableIntr

* init serial port
	D1=(5)	=IOC
	C=0	A		UART off -> reset RCS, TCS, RBR, TBR
	DAT1=C	1
	D1=D1+	10		=IRC
	DAT1=C	1		wire mode
	D1=D1-	13		=BAU
	LC(1)	#6		9600 baud
	DAT1=C	1
	D1=D1+	3		=IOC
	LC(1)	#8		SON = 1, no interrupts
	DAT1=C	1

* init bank switcher
	LC(5)	#FF000		CE1 (bank switcher), size 2KB
	CONFIG
	LC(5)	#3F000		CE1 (bank switcher), address 3F000
	CONFIG

* R3 = bank address
* R4 = 128KB counter
	ST=0	sFirst		first NAK
	ST=0	sEOT		not end of transmitting
	D=0	B		init block no.

	R3=C			save bank address
	P=	15
	LA(1)	#8		A[S] = 8 banks
	P=	0
* send next 128KB ROM bank
nBank	A=A-1	S		banks left to send
	GONC	+		yes
	ST=1	sEOT		end of transmitting
	GOTO	TxEOT		send EOT

* switch to next ROM view
+	C=R3			get next ROM view
	D0=C
	DAT0=A	B		set ROM view
	D0=D0+	2		next view
	CD0EX
	R3=C

* send 128KB block
	D0=(5)	#40000		start address of 128KB block
	LC(3)	#400		init 1024 * 128 byte = 128KB loop counter
nBlock	R4=C			save 128KB counter
	?ST=0	sFirst		wait for first NAK
	GOYES	RBYTE		receive byte

* ACK, send new block
TxBlock	ST=1	sFirst		first NAK accepted
	D=D+1	B		next block no.
	GOSUB	SBLOCK		send block

* D[S] = no of CAN reveived
* D[B] = block number
ClrCAN	ST=0	sCAN		no CAN received

* receive answer
RBYTE	D1=(2)	=RCS
-	GOSBVL	=$jCINRTN	get key
	?CBIT=1	15		ON pressed
	GOYES	FAIL1

* check for character
	A=DAT1	1		character received
	?ABIT=0	0		no
	GOYES	-

* get character
	D1=(2)	=RBR
	A=DAT1	B

* ACK
	LC(2)	ACK		ACK received?
	?A#C	B
	GOYES	noACK		no

* ACK of EOT
	?ST=0	sEOT		ACK of EOT
	GOYES	lBlock
	GOTO	EXIT		bye

* last block send?
lBlock	C=R4			get 128KB loop counter
	C=C-1	A		all 128 byte blocks send
	?C#0	X		no
	GOYES	nBlock
	GOTO	nBank		send next 128KB bank

FAIL1	GOTO	FAIL
RBYTE1	GOTO	RBYTE

* NAK
noACK	LC(2)	NAK		NAK received?
	?A#C	B
	GOYES	noNAK		no

* NAK of EOT
	?ST=0	sEOT		NAK of EOT
	GOYES	NoEOT
TxEOT	LC(2)	EOT
	GOSUB	SBYTE		send EOT
	GOTO	ClrCAN

NoEOT	?ST=0	sFirst		first NAK
	GOYES	TxBlock		send first block

* NAK, send old block
	AD0EX			D0 = data begin of old block
	LC(5)	#100
	A=A-C	A
	D0=A
	GOSUB	SBLOCK		send old block
	GOTO	ClrCAN		clean CAN marker, receive byte

* 'C'
noNAK	?ST=1	sFirst		first NAK or 'C' already received
	GOYES	noC		yes, don't check for 'C'
	LC(2)	CHAR_C		'C' received?
	?A#C	B
	GOYES	noC		no
	ST=1	sCRC		use CRC instead of checksum
	GOTO	TxBlock		send first block

* CAN
noC	LC(2)	CAN		CAN received?
	?A#C	B
	GOYES	RBYTE1		no, is garbage
* CAN, set marker
	?ST=1	sCAN		2'nd CAN ?
	GOYES	CBRK
	ST=1	sCAN		next try
	GOTO	RBYTE1

EXIT	A=0	A		return A[A] = 0
	GOTO	+

* upload break
FAIL	LC(2)	CAN		cancel
	GOSUB	SBYTE		send CAN
	GOSUB	SBYTE		send CAN
CBRK	A=0	A		return A[A] = 1
	A=A+1	A

* restore BS value
+	LC(5)	#3F000		CE1 (bank switcher)
	D0=(5)	=CurROMBank2	get last written BS address [A]
	C=DAT0	B		CE1 (bank switcher), address (3F0xxH)
	D0=C			reload BS address
	DAT0=C	B		dummy write to reload BS
	UNCNFG			unconfig CE1 (bank switcher)

* close serial port
	D1=(2)	=TCS
-	C=DAT1	1		wait for transmitter empty
	?CBIT=1	1
	GOYES	-
	D1=D1-	2		=IOC
	C=0	A		SON = 0
	DAT1=C	1

	GOSBVL	=$jAllowIntr	enable interrupts
	GOVLNG	=$jPUSH#ALOOP

* send 128 byte block
* D0   = begin of block
* D[B] = block number
SBLOCK	LC(2)	SOH		block header
	GOSUB	SBYTE
	C=D	B		block number
	GOSUB	SBYTE		send data byte
	C=-C-1  B		1' complement
	GOSUB	SBYTE		send data byte
	B=0	A		reset checksum
	LA(2)	#80		loop 128 bytes
-	C=DAT0	B		get data
	D0=D0+	2		next address
	GOSUB	SBYTE		send data byte
	?ST=1	sCRC		CRC enabled
	GOYES	+		yes, update CRC
	B=B+C	B		add checksum
	GOTO	NBYTE
+	CSRC			C[S] = low nibble byte, C[0] = high nibble byte
	GOSUB	CrcUpd		update CRC with high nibble
	CSLC			C[0] = low nibble byte
	GOSUB	CrcUpd		update CRC with low nibble
NBYTE	A=A-1	B		next byte
	?A#0	B
	GOYES	-

* send checksum or CRC
	?ST=0	sCRC		CRC disabled
	GOYES	+		skip sending high byte CRC
	C=B	A		high byte CRC
	CSR	A
	CSR	A
	GOSUB	SBYTE		send data byte
+	C=B	B		send checksum or low byte CRC, fall into SBYTE

* toggle IO annunciator
	D1=(2)	(=ANNCTRL)+1
	C=DAT1	XS
	C=C+1	XS		toggle IO annunciator bit
	C=C+1	XS
	CBIT=0  10		clear XTRA
	DAT1=C  XS

* send C[B]
SBYTE	D1=(5)	=TCS
-	C=DAT1	XS		wait for transmit buffer empty
	?CBIT=1	8
	GOYES	-
	D1=(2)	=TBR		send byte
	DAT1=C	B
	RTN

* update CCITT-16 CRC
* C[0] = data
* B[3:0] = CRC
* use D1,C[A],B[W]
CrcUpd	BSRC			B[14:13] = low byte CRC
	BSRC			B[S] = high byte low nibble CRC
	BSRC			B[0] = high byte CRC

	D1=C			C[0] = C[0] XOR B[0]
	C=C!B	P
	CD1EX
	B=B&C	P
	CD1EX
	C=C-B	P

* crc table offset
	B=0	A
	B=C	P
	B=B+B	A		CRC table offset: B[A] = B[A] * 4
	B=B+B	A

	GOSUB	+
* CCITT-16 x^16+x^12+x^5+1 Polynom table
	CON(4)	#0000
	CON(4)	#1021
	CON(4)	#2042
	CON(4)	#3063
	CON(4)	#4084
	CON(4)	#50A5
	CON(4)	#60C6
	CON(4)	#70E7
	CON(4)	#8108
	CON(4)	#9129
	CON(4)	#A14A
	CON(4)	#B16B
	CON(4)	#C18C
	CON(4)	#D1AD
	CON(4)	#E1CE
	CON(4)	#F1EF

+	C=RSTK			C[A] base address of table
	C=C+B	A
	CD1EX
	C=DAT1	4		C[3:0] CRC from table

* wCrctab ^ (wCrc << 4)
	BSLC			B[X] restored 3 lower nibble of CRC
	BSLC
	BSLC
	BSL	A		B[3:0] nibble shifted crc

	D1=C			save original C[3:0]
	C=C!B	A		B[A] = B[A] XOR C[A]
	CD1EX
	B=B&C	A
	CD1EX
	B=C-B	A		B[3:0] updated CRC
	RTN
ENDCODE

	#0=ITE
	  $ "Upload successful!"
	  $ "Upload failed!"

	FOUR TWENTYFOUR Put5x7
  ;

ASSEMBLE
endCode
RPL

ASSEMBLE
	Vfield DirCode,DirAlt
RPL
  { }

ASSEMBLE
	Vfield DirAlt,DirNot

	CON(5)  =DOCSTR
	REL(5)	+
	NIBASC  "ROM Upload\n"
 	NIBASC  "(c) by C. Giesselink\n"
 	NIBASC  "\n"
 	NIBASC  "Upload the HP39/40G\n"
 	NIBASC  "ROM to your computer\n"
 	NIBASC  "using the XModem\n"
 	NIBASC  "protocol."
+

ASSEMBLE
	Vfield DirNot,DirEnd
RPL
  { }
